<?php

namespace Services;

use Util\Utilities;
use Util\InstanceProvider;
use Util\RepositoryInstance;
use Exception\Login as Exception;
use Teknisa\Libs\Util\Utilities as LibUtilities;

class OAuth2 {
    const IS_ACTIVE_POSITIVE = 'Y';
    const DATE_TYPE_DAY = 'D';
    const ID_RESET_PASSWORD = 'R';
    const USES_ACTIVE_DIRECTORY_POSITIVE = 'Y';
    const USES_PASSWORD_CONTROL_POSITIVE = 'Y';
    const USES_IP_RESTRICTION_POSITIVE = 'Y';
    const RECURRENCE_DAYS = 7;

    const BLOCKED_PASSWORD_STATUS = 'B';
    const INVALID_PASSWORD_STATUS = 'I';

    const RESET_PASSWORD_IDENTIFICATION   = 'RESET_PASSWORD';
    const EXPIRED_PASSWORD_IDENTIFICATION = 'EXPIRED_PASSWORD';
    const PASSWORD_WILL_EXPIRE_IDENTIFICATION = 'PASSWORD_WILL_EXPIRE';
    const PASSWORD_EXPIRE_TODAY_IDENTIFICATION = 'PASSWORD_EXPIRE_TODAY';

    private $loginClientSecret;

    private $libClientId;
    private $libClientSecret;

    private $licenseExceptions = [
            22, /** Product is not available */
            23, /** Menu items not found */
            24  /** Product date expired */
    ];

    public function __construct() {
        $config = LibUtilities::getLoginConfig();
        $this->loginUrl          = $config['loginUrl'];
        $this->loginClientSecret = $config['loginClientSecret'];

        $integration = LibUtilities::getLoginIntegration();
        $this->libClientId     = $integration['clientId'];
        $this->libClientSecret = $integration['clientSecret'];
    }

    /**
     * @param $user
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @return array
     * @throws \Exception
     */
    public function login($user, $password, $productId, $requesterUrl, $attempts) {
        $userData       = InstanceProvider::getLibServiceApi()->getUserData($user, $password);
        $organizationId = $userData['USER']['ORGANIZATION_ID'];
        $userId         = $userData['USER']['ID'];

        $expiredPassword = $this->validations($userData, $password, $productId, $requesterUrl, $attempts);

        $filter = [
            'requestType' => 'FilterData',
            'filter' => [
                [
                    'name' => 'ORGANIZATION_ID',
                    'operator' => '=',
                    'value' => $organizationId,
                ],
                [
                    'name' => 'PRODUCT_ID',
                    'operator' => '=',
                    'value' => $productId,
                ]
            ],
        ];

        try {
            $loginResponse = Utilities::callWebService(null, '/productLicense', $filter, 'userData', $this->loginClientSecret, $this->loginUrl);
        } catch(\Exception $e) {
            if(in_array($e->getCode(), $this->licenseExceptions)) {
                throw $e;
            }
            $loginResponse = [
                'TOKEN' => null
            ];
        }

        $this->saveOAccTekMgr([], $organizationId, $userId, []);

        $loginResponse['TOKEN'] = InstanceProvider::getOAuth()->grantAccessToken(
            $this->libClientId,
            $this->libClientSecret,
            ['sessionId' => InstanceProvider::getSession()->getId()],
            $loginResponse['TOKEN']
        );

        $sessionData = $this->factorySessionData($userData, $loginResponse, $expiredPassword);

        $loginData = [
            'username' => $user,
            'password' => $password
        ];

        $sessionData['REFRESH_TOKEN'] = uniqid();
        InstanceProvider::getSession()->set($sessionData['REFRESH_TOKEN'], json_encode($loginData));

        return $sessionData;
    }

    /**
     * @param $userData
     * @param $loginResponse
     * @param $expiredPassword
     * @return array
     */
    private function factorySessionData($userData, $loginResponse, $expiredPassword) {
        $user = $userData['USER'];
        $userInfo = [
            'USER_ID'           => $user['ID'],
            'USER_EMAIL'        => $user['EMAIL'],
            'TOKEN'             => $loginResponse['TOKEN'],
            'REDIRECT_URL'      => $loginResponse['REDIRECT_URL'],
            'USER_DATA'         => $userData
        ];
        $encryptedSession = LibUtilities::legacyEncrypt(json_encode($userInfo));

        $expirationTime = InstanceProvider::getLibUtilities()->getTokenLifeTimeParameter();

        $sessionData = [
            'USER_ID'           => $user['ID'],
            'USER_NAME'         => $user['NAME'],
            'USER_EMAIL'        => $user['EMAIL'],

            'REDIRECT_URL'      => $loginResponse['REDIRECT_URL'],
            'TOKEN'             => $loginResponse['TOKEN'],
            'USER'              => $encryptedSession,
            'EXPIRED_PASSWORD'  => $expiredPassword,
            'LOGGED'            => true,
            'EXPIRES_IN'        => $expirationTime
        ];

        return $sessionData;
    }

    /**
     * @param $userData
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @return bool|string
     * @throws \Exception
     */
    private function validations($userData, $password, $productId, $requesterUrl, $attempts) {
        if(empty($userData) || empty($userData['USER'])) {
            Exception::invalidPassword();
        }

        $passwordStatus = $this->userValidation($userData['USER'], $productId);
        $this->validateIp($userData['IP_RESTRICTION'], LibUtilities::getCurrentIp());
        $this->validateDomain($requesterUrl);

        $expiredPassword = $this->passwordValidation($userData['USER'], $password, $userData['PASSWORD_PARAMETERS'],
            true, $attempts, $passwordStatus);

        return $expiredPassword;
    }

    /**
     * @param $user
     * @param $productId
     */
    private function userValidation($user, $productId) {
        if(!isset($user['IS_ACTIVE']) || $user['IS_ACTIVE'] !== self::IS_ACTIVE_POSITIVE) {
            Exception::userInactive();
        }

        if(empty($user['ORGANIZATION_ID'])) {
            Exception::userNotOrganization();
        }

        if (empty($productId)) {
            Exception::productIdNotFound();
        }

        $operatorParams = RepositoryInstance::getOperador()->getOperatorParams($user['ID'], $user['ORGANIZATION_ID']);
        if($operatorParams['IDSTATUSENHA'] == self::INVALID_PASSWORD_STATUS) {
            Exception::accessBlocked();
        }

        return $operatorParams['IDSTATUSENHA'];
    }

    /**
     * @param $user
     * @param $password
     * @param $passwordParams
     * @param bool $isValidateExpired
     * @param $attempts
     * @return bool|string
     */
    private function passwordValidation($user, $password, $passwordParams, $isValidateExpired = true, $attempts = null, $passwordStatus = null) {
        $isAD = $passwordParams && $passwordParams['USES_ACTIVE_DIRECTORY'] == self::USES_ACTIVE_DIRECTORY_POSITIVE;

        /** If exists attempts, this method called in login.
         *  Else, called by expired session */
        if(empty($attempts)) {
            $paramSenha = null;
        } else {
            $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($user['ORGANIZATION_ID']);
            $paramSenha    = RepositoryInstance::getParamSenha()->getParamSenhaByNrOrg($user['ORGANIZATION_ID'], $entityManager);
        }

        $this->validateWebPassword($user, $password, $isAD, $attempts, $paramSenha);

        if($passwordStatus && $passwordStatus == self::BLOCKED_PASSWORD_STATUS) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if($isValidateExpired && $passwordParams) {
            return $this->validateExpiredPassword($passwordParams, $paramSenha);
        }

        return false;
    }

    /**
     * @param $user
     * @param $loginPassword
     * @param $isAD
     * @param $attempts
     * @param $paramSenha
     * @throws \Exception
     */
    private function validateWebPassword($user, $loginPassword, $isAD, $attempts, $paramSenha) {
        $loginPasswordEncrypted = !$isAD ? LibUtilities::encrypt($loginPassword) : $loginPassword;
        if($user['PASSWORD_MD5'] != $loginPasswordEncrypted) {
            $this->throwInvalidPasswordError($user['ID'], $user['ORGANIZATION_ID'], $attempts, $paramSenha);
        }
    }

    /**
     * @param $userId
     * @param $organizationId
     * @param $attempts
     * @param $paramSenha
     * @throws \Exception
     */
    private function throwInvalidPasswordError($userId, $organizationId, $attempts, $paramSenha) {
        if(empty($paramSenha)) {
            Exception::invalidPassword();
        } else {
            $nrErrors = $paramSenha[0]['NRERROSENHA'];
            if(!empty($nrErrors) && $attempts == $nrErrors - 1) {
                Exception::invalidPasswordBlockUser();
            } else if(!empty($nrErrors) && $attempts >= $nrErrors) {
                RepositoryInstance::getOperador()->saveOperatorPasswordStatus($userId, $organizationId, self::INVALID_PASSWORD_STATUS);
                Exception::accessBlockedByAttempts();
            } else {
                Exception::invalidPasswordCode32();
            }
        }
    }

    /**
     * @param $passwordParams
     * @param $paramSenha
     * @return bool|string
     */
    private function validateExpiredPassword($passwordParams, $paramSenha) {
        if($passwordParams['IDENTIFICATION'] == self::ID_RESET_PASSWORD) {
            return self::RESET_PASSWORD_IDENTIFICATION;
        }
        if(empty($passwordParams['UPDATE_DATE'])) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if(!empty($passwordParams['PASSWORD_EXPIRATION_DAYS']) && $passwordParams['USES_PASSWORD_CONTROL'] == self::USES_PASSWORD_CONTROL_POSITIVE) {
            $expireDate = \DateTime::createFromFormat('d/m/Y', $passwordParams['UPDATE_DATE']);
            $interval = 'P' . ($passwordParams['PASSWORD_EXPIRATION_DAYS']) . self::DATE_TYPE_DAY;
            $expireDate->add(new \DateInterval($interval));
            $actualDate = new \DateTime();
            $intervalExp  = Utilities::getDateIntervalInSeconds($expireDate, $actualDate);
            $intervalDays = Utilities::getDateIntervalInSeconds($actualDate, $expireDate);
            if($intervalDays > 0) {
                $intervalDays = $intervalDays / 86400;
            } else if($intervalDays < 0) {
                $intervalDays = null;
            }
            if ($intervalExp > 0) {
                return self::EXPIRED_PASSWORD_IDENTIFICATION;
            } else if($intervalDays == 0) {
                return self::PASSWORD_EXPIRE_TODAY_IDENTIFICATION;
            } else if($intervalDays && $intervalDays <= $paramSenha[0]['NRDIASAVISO']) {
                return self::PASSWORD_WILL_EXPIRE_IDENTIFICATION . '_' . $intervalDays;
            }
        }
        return false;
    }

    /**
     * @param $ipRestriction
     * @param $ip
     * @throws \Exception
     */
    private function validateIp($ipRestriction, $ip) {
        if(!empty($ipRestriction) && $ipRestriction['USES_IP_RESTRICTION'] == self::USES_IP_RESTRICTION_POSITIVE) {
            $isValidIp = false;
            if(Utilities::isIpv4($ip)) {
                foreach($ipRestriction['IPV4'] as $ipV4) {
                    if(Utilities::validateIpInRange($ip, $ipV4)) {
                        $isValidIp = true;
                        break;
                    }
                }
            } else {
                foreach($ipRestriction['IPV6'] as $ipV6) {
                    if($ip == $ipV6) {
                        $isValidIp = true;
                        break;
                    }
                }
            }
            if(!$isValidIp) {
                Exception::ipIsOutOfRange($ip);
            }
        }
    }

    /**
     * @param $url
     * @return bool
     */
    private function validateDomain($url) {
        if(Utilities::getUseDomainFilterParameter()) {
            $validDomains = explode(';', Utilities::getDomainFilterParameter());
            foreach($validDomains as $domain) {
                if(!empty($domain) && strpos($url, $domain) !== false) {
                    return true;
                }
            };
            Exception::invalidDomain();
        }
        return true;
    }

    /**
     * @param array $productsPermission
     * @param $organizationId
     * @param $userId
     * @param $relatedOrgIds
     * @throws \Exception
     */
    private function saveOAccTekMgr($productsPermission, $organizationId, $userId, $relatedOrgIds) {
        $oAccTekMgrRepository = RepositoryInstance::getOAccTekMgr();
        $licences   = ['products' => $productsPermission, 'macAddress' => Utilities::getMacAddress(), 'relatedOrgIds' => $relatedOrgIds];
        $licences   = LibUtilities::legacyEncrypt(json_encode($licences), LibUtilities::LOGIN_CRYPT_STRING);
        $oAccTekMgr = $oAccTekMgrRepository->getOAccTekMgrByNrOrg($organizationId);

        if(empty($oAccTekMgr)) {
            $oAccTekMgrRepository->insertOAccTekMgr($organizationId, $userId, self::RECURRENCE_DAYS, $licences);
        } else {
            $lastupdate = \DateTime::createFromFormat('d/m/Y', $oAccTekMgr['DTULTATU']);
            $interval   = 'P' . $oAccTekMgr['RECURRENCE_DAYS'] . self::DATE_TYPE_DAY;
            $lastupdate->add(new \DateInterval($interval));
            $actualDate = new \DateTime();

            $oAccTek = @json_decode(LibUtilities::legacyDecrypt($oAccTekMgr['OACCTEK'], LibUtilities::LOGIN_CRYPT_STRING), true);
            $actualRelatedOrgIds = isset($oAccTek['relatedOrgIds']) ? $oAccTek['relatedOrgIds'] : [];
            $isDiffOrgIds = !empty(array_diff($relatedOrgIds, $actualRelatedOrgIds)) || !empty(array_diff($actualRelatedOrgIds, $relatedOrgIds));

            if (Utilities::getDateIntervalInSeconds($lastupdate, $actualDate) > 0 || $isDiffOrgIds) {
                $oAccTekMgrRepository->updateOAccTekMgr($organizationId, $userId, self::RECURRENCE_DAYS, $licences);
            }
        }
    }
}